""".. Ignore pydocstyle D400.

=======================
Auditlog Configuration
=======================

"""

from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class FlowConfig(AppConfig):
    """Audit log AppConfig."""

    name = "resolwe.auditlog"
    verbose_name = _("Resolwe Auditlog")



# Generated by Django 4.2.2 on 2023-09-14 14:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0014_alter_annotationfield_validator_regex"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationfield",
            name="vocabulary",
            field=models.JSONField(blank=True, null=True),
        ),
    ]



"""Logging configuration for executors."""

import asyncio
import json
import logging
import os
import socket

from .global_settings import DATA_ID


class JSONFormatter(logging.Formatter):
    """Logger formatter for dumping records to JSON."""

    def format(self, record):
        """Dump the record to JSON."""
        data = record.__dict__.copy()

        data["data_id"] = DATA_ID
        data["hostname"] = socket.gethostname()

        # Get relative path, so listener can reconstruct the path to the actual code.
        data["pathname"] = os.path.relpath(data["pathname"], os.path.dirname(__file__))

        # Exception and Traceback cannot be serialized.
        data["exc_info"] = None

        return json.dumps(data, default=str)


class ZeromqHandler(logging.Handler):
    """Publish messages to Redis channel."""

    def __init__(self, communicator, **kwargs):
        """Construct a handler instance."""
        self.communicator = communicator
        self.loop = asyncio.get_event_loop()
        super().__init__(**kwargs)

    def emit(self, record):
        """Send log message to the listener.

        Log all possible exceptions as warnings.
        """
        try:
            asyncio.run_coroutine_threadsafe(
                self.communicator.log(self.format(record)), self.loop
            )
        except Exception as e:
            # Not really much we can do here. Just print exception details to stderr.
            print(e)


def configure_logging(communicator):
    """Configure logging to send log records to the master."""
    # Get and configure root logger.
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    # Create and configure zmq handler.
    zmq_handler = ZeromqHandler(communicator)
    zmq_handler.setLevel(logging.DEBUG)
    zmq_handler.setFormatter(JSONFormatter())

    # Create and configure console_handler.
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.WARNING)

    # Add zeromq and console handlers to root logger.
    root_logger.addHandler(zmq_handler)
    root_logger.addHandler(console_handler)



""".. Ignore pydocstyle D400.

=====
State
=====

Constants used by the dispatcher.
.. autofunction:: resolwe.flow.managers.state.update_constants
"""

# This module should not import anything local, or there will be circular
# dependencies, since the constants are needed in various sub-modules inside
# resolwe.flow.managers.

from collections import namedtuple

from django.conf import settings

ManagerChannelPair = namedtuple("ManagerChannelPair", ["queue", "queue_response"])

MANAGER_CONTROL_CHANNEL = "DUMMY.control"
LISTENER_CONTROL_CHANNEL = "DUMMY.listener"


def update_constants():
    """Recreate channel name constants with changed settings.

    This kludge is mostly needed due to the way Django settings are
    patched for testing and how modules need to be imported throughout
    the project. On import time, settings are not patched yet, but some
    of the code needs static values immediately. Updating functions such
    as this one are then needed to fix dummy values.
    """
    global MANAGER_CONTROL_CHANNEL, LISTENER_CONTROL_CHANNEL
    redis_prefix = getattr(settings, "FLOW_MANAGER", {}).get("REDIS_PREFIX", "")

    MANAGER_CONTROL_CHANNEL = "{}.control".format(redis_prefix)
    LISTENER_CONTROL_CHANNEL = "{}.listener".format(redis_prefix)


update_constants()



# Generated by Django 4.2.7 on 2023-12-14 07:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0019_alter_annotationpreset_modified_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationvalue",
            name="field",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="values",
                to="flow.annotationfield",
            ),
        ),
    ]



# Generated by Django 4.2.7 on 2023-12-06 11:59

from django.db import migrations

import resolwe.flow.models.base


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0018_add_annotationvalue_default_order"),
    ]

    operations = [
        migrations.AlterField(
            model_name="annotationpreset",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="collection",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="data",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="descriptorschema",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="entity",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="process",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="relation",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="storage",
            name="modified",
            field=resolwe.flow.models.base.ModifiedField(auto_now=True, db_index=True),
        ),
    ]



"""Iterator utils."""

import collections


def iterate_fields(fields, schema, path_prefix=None):
    """Iterate over all field values sub-fields.

    This will iterate over all field values. Some fields defined in the schema
    might not be visited.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :return: (field schema, field value)
    :rtype: tuple

    """
    if path_prefix is not None and path_prefix != "" and path_prefix[-1] != ".":
        path_prefix += "."

    schema_dict = {val["name"]: val for val in schema}
    for field_id, properties in fields.items():
        path = "{}{}".format(path_prefix, field_id) if path_prefix is not None else None
        if field_id not in schema_dict:
            raise KeyError("Field definition ({}) missing in schema".format(field_id))
        if "group" in schema_dict[field_id]:
            for rvals in iterate_fields(
                properties, schema_dict[field_id]["group"], path
            ):
                yield rvals if path_prefix is not None else rvals[:2]
        else:
            rvals = (schema_dict[field_id], fields, path)
            yield rvals if path_prefix is not None else rvals[:2]


def iterate_schema(fields, schema, path_prefix="", include_groups=False):
    """Iterate over all schema sub-fields.

    This will iterate over all field definitions in the schema. Some field v
    alues might be None.

    :param fields: field values to iterate over
    :type fields: dict
    :param schema: schema to iterate over
    :type schema: dict
    :param path_prefix: dot separated path prefix
    :type path_prefix: str
    :return: (field schema, field value, field path)
    :rtype: tuple

    """
    if path_prefix and path_prefix[-1] != ".":
        path_prefix += "."

    for field_schema in schema:
        name = field_schema["name"]
        if "group" in field_schema:
            if include_groups:
                yield (field_schema, fields, "{}{}".format(path_prefix, name))
            for rvals in iterate_schema(
                fields[name] if name in fields else {},
                field_schema["group"],
                "{}{}".format(path_prefix, name),
            ):
                yield rvals
        else:
            yield (field_schema, fields, "{}{}".format(path_prefix, name))


def iterate_dict(container, exclude=None, path=None):
    """Iterate over a nested dictionary.

    The dictionary is iterated over in a depth first manner.

    :param container: Dictionary to iterate over
    :param exclude: Optional callable, which is given key and value as
        arguments and may return True to stop iteration of that branch
    :return: (path, key, value) tuple
    """
    if path is None:
        path = []

    for key, value in container.items():
        if callable(exclude) and exclude(key, value):
            continue

        if isinstance(value, collections.Mapping):
            for inner_path, inner_key, inner_value in iterate_dict(
                value, exclude=exclude, path=path + [key]
            ):
                yield inner_path, inner_key, inner_value

        yield path, key, value



"""Resolwe process serializer."""

from resolwe.flow.models import Process

from .base import ResolweBaseSerializer


class ProcessSerializer(ResolweBaseSerializer):
    """Serializer for Process objects."""

    class Meta:
        """ProcessSerializer Meta options."""

        model = Process
        read_only_fields = (
            "created",
            "id",
            "modified",
            "is_active",
        )
        update_protected_fields = (
            "category",
            "contributor",
            "data_name",
            "description",
            "entity_always_create",
            "entity_input",
            "entity_type",
            "input_schema",
            "name",
            "output_schema",
            "persistence",
            "requirements",
            "run",
            "scheduling_class",
            "slug",
            "type",
            "version",
        )
        fields = read_only_fields + update_protected_fields



"""Resolwe models utils."""

from resolwe.flow.utils import iterate_fields

from .duplicate import (  # noqa: F401
    bulk_duplicate_collection,
    bulk_duplicate_data,
    bulk_duplicate_entity,
)
from .hydrate import (  # noqa: F401
    hydrate_input_references,
    hydrate_input_uploads,
    hydrate_size,
)
from .reference import referenced_files  # noqa: F401
from .render import (  # noqa: F401
    fill_with_defaults,
    json_path_components,
    render_descriptor,
    render_template,
)
from .validation import (  # noqa: F401
    DirtyError,
    validate_data_object,
    validate_process_types,
    validate_schema,
    validation_schema,
)


def get_collection_of_input_entities(data):
    """Get collection that contains all "entity inputs" of a given data.

    With "entity input", one refers to the inputs that are part of an entity.
    """
    # Prevent circular imports:
    from resolwe.flow.models import Collection

    data_ids = set()

    for field_schema, fields in iterate_fields(data.input, data.process.input_schema):
        name = field_schema["name"]
        value = fields[name]
        if "type" not in field_schema:
            continue

        if field_schema["type"].startswith("data:"):
            value = [value]
        elif not field_schema["type"].startswith("list:data:"):
            continue

        data_ids.update([val for val in value if val is not None])

    collections = Collection.objects.filter(
        data__in=list(data_ids),
        data__entity__isnull=False,
    ).distinct()

    if collections.count() != 1:
        raise ValueError(
            "Entity inputs should be part of exactly one collection. (not {})".format(
                len(collections)
            )
        )

    return collections.first()


def serialize_collection_relations(collection):
    """Serialize all relations of a collection."""
    # Prevent circular imports:
    from resolwe.flow.models import Relation, RelationPartition

    serialized_relations = []
    for relation in Relation.objects.filter(collection=collection):
        serialized_relations.append(
            {
                "relation_id": relation.id,
                "relation_type_name": relation.type.name,
                "relation_type_ordered": relation.type.ordered,
                "category": relation.category,
                "unit": relation.unit,
                "partitions": list(
                    RelationPartition.objects.filter(relation=relation).values()
                ),
            }
        )

    return serialized_relations



"""Resolwe models render utils."""

from resolwe.flow.utils import dict_dot, iterate_schema


def render_descriptor(data):
    """Render data descriptor.

    The rendering is based on descriptor schema and input context.

    :param data: data instance
    :type data: :class:`resolwe.flow.models.Data` or :class:`dict`

    """
    if not data.descriptor_schema:
        return

    # Set default values
    for field_schema, field, path in iterate_schema(
        data.descriptor, data.descriptor_schema.schema, "descriptor"
    ):
        if "default" in field_schema and field_schema["name"] not in field:
            dict_dot(data, path, field_schema["default"])


def render_template(process, template_string, context):
    """Render template using the specified expression engine."""
    from resolwe.flow.managers import manager

    # Get the appropriate expression engine. If none is defined, do not evaluate
    # any expressions.
    expression_engine = process.requirements.get("expression-engine", None)
    if not expression_engine:
        return template_string

    return manager.get_expression_engine(expression_engine).evaluate_block(
        template_string, context
    )


def json_path_components(path):
    """Convert JSON path to individual path components.

    :param path: JSON path, which can be either an iterable of path
        components or a dot-separated string
    :return: A list of path components
    """
    if isinstance(path, str):
        path = path.split(".")

    return list(path)


def fill_with_defaults(process_input, input_schema):
    """Fill empty optional fields in input with default values."""
    for field_schema, fields, path in iterate_schema(
        process_input, input_schema, include_groups=True
    ):
        if "group" in field_schema and field_schema["name"] not in fields:
            dict_dot(process_input, path, {})
        if "default" in field_schema and field_schema["name"] not in fields:
            dict_dot(process_input, path, field_schema["default"])



""".. Ignore pydocstyle D400.

=====================
Permissions shortcuts
=====================

.. autofunction:: _group_groups
.. autofunction:: get_object_perms

"""

from typing import Dict, List, Optional, Tuple

from django.contrib.auth.models import Group, User
from django.db import models

from resolwe.permissions.models import (
    Permission,
    PermissionList,
    UserOrGroup,
    get_anonymous_user,
)
from resolwe.permissions.utils import get_identity, model_has_permissions


def get_user_group_perms(
    user_or_group: UserOrGroup, obj: models.Model
) -> Tuple[PermissionList, List[Tuple[int, str, PermissionList]]]:
    """Get permissions for the user/group on the given model.

    This method is only used in Resolwe views and expects permissions to be
    prefetched by the view. To avoid additional database hits, the filtering is
    done in Python.

    If the given user_or_group is user with superuser privileges, only directly
    attached permissions are returned.

    :returns: the tuple cosisting of permissions for the given user and
    a list of permissions for every group the user belongs to.
    """
    if not model_has_permissions(obj):
        return [], []

    entity, entity_name = get_identity(user_or_group)
    user_perms = []
    groups_perms = []
    if entity_name == "user":
        if not entity.is_active:
            return [], []

        for permission_model in obj.permission_group.permissions.all():
            if permission_model.user_id == entity.id:
                user_perms = permission_model.permissions

    # The correct group permissions are pre-fetched, iterate over them.
    for permission_model in obj.permission_group.permissions.all():
        if permission_model.group_id is not None:
            groups_perms.append(
                (
                    permission_model.group_id,
                    permission_model.group.name,
                    permission_model.permissions,
                )
            )
    return user_perms, groups_perms


def _get_users_with_perms(obj: models.Model) -> Dict[User, PermissionList]:
    """Get users with permissions on the given object.

    Only permissions attached to users directly are considered.

    The users with administrative privileges without permissions attached
    directly to the given object are not returned, even though they have
    implicit 'owner' permissions.

    When user with administrative privileges has permissions attached to a
    given object the permission in the database is returned (and not the
    owner permission).

    This method is only used in Resolwe views and expects permissions to be
    prefetched by the view. To avoid additional database hits, the filtering is
    done in Python.
    """
    return {
        permission_model.user: permission_model.permissions
        for permission_model in obj.permission_group.permissions.all()
        if permission_model.user_id is not None and permission_model.value > 0
    }


def _get_groups_with_perms(obj: models.Model) -> Dict[Group, PermissionList]:
    """Get groups with permissions on the given object.

    This method is only used in Resolwe views and expects permissions to be
    prefetched by the view. To avoid additional database hits, the filtering is
    done in Python.
    """
    return {
        permission_model.group: permission_model.permissions
        for permission_model in obj.permission_group.permissions.all()
        if permission_model.group_id is not None and permission_model.value > 0
    }


def get_object_perms(
    obj: models.Model,
    user: Optional[User] = None,
    mock_superuser_permissions: bool = False,
) -> List[Dict]:
    """Return permissions for given object in Resolwe specific format.

    Function returns permissions for given object ``obj`` in the following
    format::

       {
           "type": "group"/"user"/"public",
           "id": <group_or_user_id>,
           "name": <group_or_user_name>,
           "username": <username>,
           "permissions": [<first_permission>, <second_permission>,...]
       }

    For ``public`` type ``id`` and ``name`` keys are omitted.

    If ``user`` parameter is given, permissions are limited only to
    given user, groups he belongs to and public permissions.

    This function should be only used from Resolwe views: since permissions
    for the current user (users when user has share permission on the given
    object) are prefetched, we only iterate through objects here and filter
    them in Python. Using filter method would result in a new database query.

    :param obj: Resolwe's DB model's instance
    :param user: Django user
    :param mock_superuser_permissions: when True return all permissions for
        users that are superusers
    :return: list of permissions object in described format

    """
    perms_list = []
    public_permissions = []
    anonymous_user = get_anonymous_user()

    # Request only permissions for the given user.
    if user is not None:
        if user.is_authenticated:
            user_permissions, groups_permissions = get_user_group_perms(user, obj)
            if mock_superuser_permissions and user.is_superuser:
                user_permissions = list(Permission.highest())
        else:
            user_permissions, groups_permissions = [], []

        if user_permissions:
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": [str(permission) for permission in user_permissions],
                }
            )

        if groups_permissions:
            for group_id, group_name, permissions in groups_permissions:
                perms_list.append(
                    {
                        "type": "group",
                        "id": group_id,
                        "name": group_name,
                        "permissions": [str(permission) for permission in permissions],
                    }
                )
    else:
        for user, permissions in _get_users_with_perms(obj).items():
            if user == anonymous_user:
                # public user permissions are considered bellow
                public_permissions = permissions
                continue
            perms_list.append(
                {
                    "type": "user",
                    "id": user.pk,
                    "name": user.get_full_name() or user.username,
                    "username": user.username,
                    "permissions": [str(permission) for permission in permissions],
                }
            )

        for group, permissions in _get_groups_with_perms(obj).items():
            perms_list.append(
                {
                    "type": "group",
                    "id": group.pk,
                    "name": group.name,
                    "permissions": [str(permission) for permission in permissions],
                }
            )
    # Retrieve public permissions only if they are not set.
    # if not public_permissions:
    public_permissions = get_user_group_perms(anonymous_user, obj)[0]

    if public_permissions:
        perms_list.append(
            {
                "type": "public",
                "permissions": [str(permission) for permission in public_permissions],
            }
        )

    return perms_list



"""Resolwe models duplicate utils."""

from copy import deepcopy
from typing import TYPE_CHECKING, Callable, Iterable, Optional, Sequence, Type, TypeVar

from django.core.exceptions import ValidationError
from django.db import IntegrityError, models, transaction
from django.utils import timezone

from resolwe.flow.models import AnnotationField, AnnotationValue
from resolwe.flow.utils import iterate_fields
from resolwe.permissions.models import Permission, PermissionGroup, PermissionObject
from resolwe.permissions.utils import assign_contributor_permissions, copy_permissions

if TYPE_CHECKING:
    from resolwe.flow.models import Collection, Data, Entity
    from resolwe.flow.models.collection import CollectionQuerySet
    from resolwe.flow.models.data import DataQuerySet
    from resolwe.flow.models.entity import EntityQuerySet

DataPK = int
DataPKMap = dict[DataPK, DataPK]
CollectionPK = int
CollectionPKMap = dict[CollectionPK, "Collection"]
EntityPK = int
EntityPKMap = dict[EntityPK, "Entity"]


T = TypeVar("T", "Data", "Collection", "Entity")
U = TypeVar("U", bound=models.Model)


def send_duplicate_signal(
    model: Type[models.Model],
    new_objects: Sequence[models.Model],
    objects: Sequence[models.Model],
):
    """Send post_duplicate signal."""
    from resolwe.flow.signals import post_duplicate

    post_duplicate.send(sender=model, instances=new_objects, old_instances=objects)


def _rewire_data_relations(
    old_data: Sequence["Data"], new_data: Sequence["Data"]
) -> None:
    """Rewire the relations on the new objects based on old ones."""

    # Prevent circular import.
    from resolwe.flow.models import Data, DataDependency, DataMigrationHistory

    pk_mapping: dict[int, int] = {
        old.pk: new.pk for old, new in zip(old_data, new_data)
    }
    _rewire_inputs(new_data, pk_mapping)

    # Add m2m storage relations.
    through_model = Data.storages.through  # type: ignore
    storage_relations = through_model.objects.filter(data__in=old_data)

    _rewire_foreign_key(storage_relations, Data, pk_mapping)

    # Copy data dependencies.
    # We only inherit parent dependencies to keep track of the Data object's
    # source. Child dependencies reflect later actions that don't directly
    # depend on the current object, so copying them would only make unnecessary
    # noise. Duplicate dependencies are not copied to keep a simple chain with
    # links only to direct predecessors.
    data_dependencies = DataDependency.objects.filter(
        kind__in=[DataDependency.KIND_IO, DataDependency.KIND_SUBPROCESS],
        child__in=old_data,
    )
    _rewire_foreign_key(data_dependencies, Data, pk_mapping)

    # Add duplicate dependencies.
    DataDependency.objects.bulk_create(
        DataDependency(parent=old, child=new, kind=DataDependency.KIND_DUPLICATE)
        for old, new in zip(old_data, new_data)
    )

    # Copy migration history. Ordering is needed to keep the order of migrations.
    migration_history = DataMigrationHistory.objects.filter(data__in=old_data).order_by(
        "pk"
    )
    _rewire_foreign_key(migration_history, Data, pk_mapping)


def _rewire_foreign_key(
    objects: models.QuerySet, target: Type[models.Model], pk_mapping: DataPKMap
) -> list:
    """Make a copy of given objects and rewire references to target objects.

    All fields that points to an object of ``target`` type are replaced with
    the new objects according to the provided mapping. The list of new objects
    is returned. The new objects are already saved to the database.

    :param objects: A list of objects to process.
    :type objects: list of `~django.db.models.Model`

    :param target: A class to which references should be replaced
    :type terget: `~django.db.models.Model`

    :param dict pk_mapping: Dictionary with old primary keys as keys and new
        primary keys as values.

    :rtype: list of `~django.db.models.Model`

    """
    if not objects:
        return []

    # Check that all objects are instances of the same class.
    obj_class = type(objects[0])
    if any(type(obj) != obj_class for obj in objects):
        raise ValueError("All objects must be instances of the same class.")

    # Find fields pointing to the target model.
    to_map = [
        field.name
        for field in obj_class._meta.fields
        if type(field) == models.ForeignKey and field.related_model == target
    ]

    # Rewire references.
    new_objects = []
    for obj in objects:
        new_obj = deepcopy(obj)
        for field_name in to_map:
            field_name_id = f"{field_name}_id"
            if value := pk_mapping.get(getattr(obj, field_name_id)):
                setattr(new_obj, field_name_id, value)

        new_obj.pk = None
        new_objects.append(new_obj)

    return obj_class.objects.bulk_create(new_objects)


def _rewire_inputs(data: Iterable["Data"], pk_mapping: dict[int, int]) -> None:
    """Rewire inputs of provided data objects.

    References to input data objects in ``input`` field of the Data objects are changed
    with references of their copies according to the provided mapping. If primary key
    is not in the mapping, original value is kept.

    The data objects are saved to the database using bulk_update method so no signal is
    triggered.

    :param list data: A list of Data objects to process.

    :param dict pk_mapping: A dict where keys are primary keys of original Data objects
        and values are primary keys of their copies.

    """
    from resolwe.flow.models import Data

    for datum in data:
        for field_schema, fields in iterate_fields(
            datum.input, datum.process.input_schema
        ):
            name = field_schema["name"]
            value = fields[name]

            if field_schema["type"].startswith("data:") and value in pk_mapping:
                fields[name] = pk_mapping[value]

            elif field_schema["type"].startswith("list:data:"):
                fields[name] = [
                    pk_mapping[pk] if pk in pk_mapping else pk for pk in value
                ]
    Data.objects.bulk_update(data, ["input"])


def _check_permissions(
    objects: Iterable[PermissionObject], permission: Permission, user
) -> None:
    """Check that user has permission on all given objects.

    :param objects: An iterable of objects to be checked.
    :param permission: the required permission.
    :raises: ValidationError when user does not have all the required
        permissions.
    """
    # The permissions are the same for objects in the same permission groups.
    # So only check permissions for objects belonging to the different
    # permission groups.
    # It does not matter which object for the permission group we check so we
    # pick out the last one.

    for obj in {obj.permission_group_id: obj for obj in objects}.values():
        if not user.has_perm(permission, obj):
            raise ValidationError(
                f"User doesn't have '{permission}' permission on {obj}."
            )


def _check_created_objects_permissions(
    new_data: Iterable["Data"],
    new_entities: Iterable["Entity"],
    inherit_collection: bool,
    inherit_entity: bool,
    contributor,
):
    """Check that new objects can be created in the collections."""

    from resolwe.flow.models import Collection, Entity

    if inherit_collection:
        data_collections = Collection.objects.filter(data__in=new_data).distinct()
        _check_permissions(data_collections, Permission.EDIT, contributor)
        if new_entities:
            entity_collections = (
                Collection.objects.filter(entity__in=new_entities)
                .exclude(data__in=new_data)
                .distinct()
            )
            _check_permissions(entity_collections, Permission.EDIT, contributor)

    if inherit_entity:
        data_entities = Entity.objects.filter(data__in=new_data).distinct()
        _check_permissions(data_entities, Permission.EDIT, contributor)


def process_entity(
    inherit_collection: bool, collection_mapping: CollectionPKMap
) -> Callable[["Entity"], "Entity"]:
    """Return a function that process a duplicated Entity before saving.

    :param inherit_collection: Indicates whether the collection should be inherited or
        not.

    :param collection_mapping: A dict where keys are primary keys of original
        collections and values are primary keys of their copies.
    """

    def processor(entity: "Entity") -> "Entity":
        """Set collection of given Entity.

        Collection is set to ``None`` if ``inherit_collection`` value is set to
        ``False``, or mapped according to the ``collection_mapping`` otherwise (or left
        as is if value is not in the mapping).

        :param entity: An entity to be processed.
        """
        if inherit_collection and entity.collection:
            if entity.collection.id in collection_mapping:
                entity.collection = collection_mapping[entity.collection.pk]
        else:
            entity.collection = None

        return entity

    return processor


def process_data(
    inherit_collection: bool,
    inherit_entity: bool,
    collection_mapping: CollectionPKMap = dict(),
    entity_mapping: EntityPKMap = dict(),
) -> Callable[["Data"], "Data"]:
    """Return a function that process a duplicated Data object before saving.

    :param inherit_collection: Indicates whether the collection should be inherited.

    :param inherit_entity: Indicates whether the entity should be inherited.

    :param collection_mapping: A dict where keys are primary keys of original
        collections and values are primary keys of their copies.

    :param entity_mapping: A dict where keys are primary keys of original entities and
        values are primary keys of their copies.
    """

    def processor(datum: "Data") -> "Data":
        """Set collection and entity of given Data object.

        Collection and entity are set to ``None`` if ``inherit_collection`` and
        ``inherit_entity`` values are set to ``False``, respectively, and mapped
        according to ``collection_mapping`` and ``entity_mapping`` otherwise (or left
        as is if value is not in the mapping).

        :param datum: An Data object to be processed.
        """
        if inherit_collection and datum.collection:
            if datum.collection.id in collection_mapping:
                datum.collection = collection_mapping[datum.collection.pk]
        else:
            datum.collection = None

        if inherit_entity and datum.entity:
            if datum.entity.id in entity_mapping:
                datum.entity = entity_mapping[datum.entity.pk]
        else:
            datum.entity = None

        return datum

    return processor


def copy_objects(
    objects: Sequence[T],
    contributor,
    name_prefix: str,
    obj_processor: Optional[Callable[[T], T]] = None,
    send_signals: bool = True,
) -> list[T]:
    """Make a copy of given queryset.

    Shallow copy given queryset and set contributor to the given value, prepend name
    with the prefix, set slug to a unique value, and set ``duplicated`` date to the
    current time. Special attention is paid to keep the ``created`` date to its
    original value.

    If ``obj_processor`` function is given, each object is passed to it and its return
    value is used instead of it.

    :param objects: A queryset to be copied.

    :param contributor: A Django user that will be assigned to copied objects as
        contributor.

    :param name_prefix: A prefix that will be prepend to the name of copied objects.

    :param send_signals: Should a post_duplicate signal be sent after copying objects.
    """

    if not objects:
        return []

    name_prefix = name_prefix or "Copy of"

    first = objects[0]
    name_max_length: int = first._meta.get_field("name").max_length  # type: ignore
    model = first._meta.model

    new_objects: list[T] = []
    for obj in objects:
        new_obj = deepcopy(obj)
        new_obj.pk = None
        new_obj.slug = None
        new_obj.contributor = contributor
        new_obj.name = f"{name_prefix} {obj.name}"
        # Truncate name if necessary.
        if len(new_obj.name) > name_max_length:
            # Truncate to name_max_length-3 and pad with dots until name_max_length.
            new_obj.name = f"{new_obj.name:.<{name_max_length}.{name_max_length-3}}"

        if obj_processor:
            new_obj = obj_processor(new_obj)

        new_objects.append(new_obj)

    try:
        # Add another atomic block to avoid corupting the main one.
        with transaction.atomic():
            model.objects.bulk_create(new_objects)
    except IntegrityError:
        # Probably a slug collision occured, try to create objects one by one.
        for obj in new_objects:
            obj.slug = None
            # Call the bulk_create to skip signal sending.
            model.objects.bulk_create([obj])

    # Send the signal after the transaction is commited.
    if send_signals:
        transaction.on_commit(
            lambda: send_duplicate_signal(model, new_objects, objects)
        )

    object_permission_group = dict()
    not_in_container = list()
    for old, new in zip(objects, new_objects):
        new.created = old.created
        new.duplicated = timezone.now()

        # Deal with permissions. When object is in container fix the pointer
        # to permission_group object.
        # When object is not in container new PermissionGroup proxy object must
        # be created, assigned to new object and permissions copied from old
        # object to new one.
        if (topmost_container := new.topmost_container) is not None:
            new.permission_group = topmost_container.permission_group
        else:
            not_in_container.append((new, old))
            object_permission_group[new] = PermissionGroup()

    PermissionGroup.objects.bulk_create(object_permission_group.values())
    for new, old in not_in_container:
        new.permission_group = object_permission_group[new]
        copy_permissions(old, new)
        assign_contributor_permissions(new, contributor)

    model.objects.bulk_update(
        new_objects, ["created", "duplicated", "permission_group"]
    )
    return new_objects


@transaction.atomic
def bulk_duplicate_collection(
    collections: "CollectionQuerySet",
    contributor,
    name_prefix: str = "",
) -> list["Collection"]:
    """Make a copy of given collection queryset.

    The contained objects (Entities and Data) are also copied.

    Copied objects are transformed in the following ways:

    * ``name_prefix`` ("Copy of " by default) string is prepend to names of all
      copied objects
    * all contained objects are attached to new collections and entities
    * ``input`` fields of all copied Data objects are processed and all inputs
      are replaced with their copies if they exist
    * permissions are copied from original objects
    * Data migration history is copied and linked to the new Data objects

    :param collections: A collection queryset to duplicate.
    :param contributor: A Django user that will be assigned to copied objects
        as contributor.
    :param name_prefix: A prefix that will be prepend to the name of all
        copied objects.
    """
    # Prevent circular imports.
    from resolwe.flow.models import Collection, Entity

    if not isinstance(collections, models.QuerySet):
        raise ValueError("Function only supports duplicating querysets.")

    # Make a list from collections to make sure the sort order is respected.
    # The queryset is evaluated anyway so this is not a performance issue.
    collection_list = list(collections)

    new_collections: list[Collection] = copy_objects(
        collection_list, contributor, name_prefix
    )

    collection_mapping: CollectionPKMap = dict()
    for old, new in zip(collection_list, new_collections):
        collection_mapping[old.pk] = new
        AnnotationField.add_to_collection(old, new)

    bulk_duplicate_entity(
        Entity.objects.filter(collection__in=collection_list),
        True,
        name_prefix,
        contributor,
        collection_mapping,
    )
    """"""
    # The data in the entity is already copied.
    return new_collections


@transaction.atomic
def bulk_duplicate_entity(
    entities: "EntityQuerySet",
    inherit_collection: bool = False,
    name_prefix: str = "",
    contributor=None,
    collection_mapping: "CollectionPKMap" = dict(),
) -> list["Entity"]:
    """Make a copy of given entity queryset.

    The contained Data objects are also copied.

    Copied objects are transformed in the following ways:

    * ``name_prefix`` ("Copy of " by default) string is prepend to names of all
      copied objects
    * Collections are preserved only if ``inherit_collection``is set to ``True``
    * all contained objects are attached to new collections and entities
    * ``input`` fields of all copied Data objects are processed and all inputs
      are replaced with their copies if they exist
    * permissions are copied from original objects
    * Data migration history is copied and linked to the new Data objects

    :param entities: An entity queryset to duplicate.
    :param contributor: A Django user that will be assigned to copied objects
        as contributor.
    :param inherit_collection: Indicates whether copied entities and data
        objects are added to the same collection as originals or not.
    :param name_prefix: A prefix that will be prepend to the name of all
        copied objects.
    """

    from resolwe.flow.models import Data

    entity_list = list(entities)
    entity_processor = process_entity(inherit_collection, collection_mapping)
    new_entities = copy_objects(entity_list, contributor, name_prefix, entity_processor)
    # Copy the existing annotations to the new entity and create mapping
    # between old and new entities.
    new_annotations = []
    for old_entity, new_entity in zip(
        entity_list, new_entities
    ):  # If collection is not set the annotations cannot be copied.
        if inherit_collection:
            new_annotations += old_entity.copy_annotations(new_entity, contributor)
        if old_entity.collection and new_entity.collection:
            new_entity.collection.annotation_fields.add(
                *old_entity.collection.annotation_fields.all()
            )
    AnnotationValue.objects.bulk_create(new_annotations)

    # Check permissions when collection is inherited.
    if inherit_collection:
        _check_created_objects_permissions([], new_entities, True, False, contributor)

    entity_mapping: EntityPKMap = {
        old.pk: new for old, new in zip(entity_list, new_entities)
    }

    bulk_duplicate_data(
        Data.objects.filter(entity__in=entity_list),
        inherit_collection,
        True,
        name_prefix,
        contributor,
        collection_mapping,
        entity_mapping,
    )
    return new_entities


@transaction.atomic
def bulk_duplicate_data(
    data: "DataQuerySet",
    inherit_collection: bool = False,
    inherit_entity: bool = False,
    name_prefix: str = "",
    contributor=None,
    collection_mapping: CollectionPKMap = dict(),
    entity_mapping: EntityPKMap = dict(),
) -> list["Data"]:
    """Make a copy of given data queryset.

    Copied objects are transformed in the following ways:

    * ``name_prefix`` ("Copy of " by default) string is prepend to names of all
      copied objects
    * Collection and/or entity of top-most copied objects are preserved only if
      ``inherit_collection`` and/or ``inherit_entity`` values are set to ``True``
    * ``input`` fields of all copied Data objects are processed and all inputs
      are replaced with their copies if they exist
    * permissions are copied from original objects
    * Data migration history is copied and linked to the new Data objects

    :param data: A data queryset to duplicate.
    :param contributor: A Django user that will be assigned to copied objects
        as contributor.
    :param inherit_collection: Indicates whether copied entities and data objects are
        added to the same collection as originals or not.
    :param inherit_entity: Indicates whether copied data  objects are added to the same
        collection as originals or not.
    :param name_prefix: A prefix that is prepended to the name of all copied objects.
    """
    from resolwe.flow.models import Data

    if data.exclude(status__in=[Data.STATUS_DONE, Data.STATUS_ERROR]).exists():
        raise ValidationError(
            "Data object must have done or error status to be duplicated."
        )
    data_list = list(data)
    data_processor = process_data(
        inherit_collection, inherit_entity, collection_mapping, entity_mapping
    )
    new_data = copy_objects(data_list, contributor, name_prefix, data_processor)
    _rewire_data_relations(data_list, new_data)
    _check_created_objects_permissions(
        new_data, [], inherit_collection, inherit_entity, contributor
    )
    return new_data



"""Thread-safe singleton class."""

from os import getpid
from threading import Lock


class Singleton:
    """Class containing the callback for zmq authentication."""

    _instance = None
    _instance_lock = Lock()
    _instance_pid: int | None = None

    @classmethod
    def has_instance(cls):
        """Check if the instance exists."""
        return not (cls._instance is None or cls._instance_pid != getpid())

    @classmethod
    def instance(cls, *args):
        """Return a global instance of the class."""
        if not cls.has_instance():
            with cls._instance_lock:
                if not cls.has_instance():
                    cls._instance = cls(*args)
                    cls._instance_pid = getpid()
        return cls._instance



# pylint: disable=missing-docstring
import os
from io import StringIO

import yaml
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase as DjangoTestCase
from django.test import override_settings

from resolwe.flow.models import Data, Process
from resolwe.permissions.models import Permission
from resolwe.test import ProcessTestCase, TestCase

PROCESSES_DIR = os.path.join(os.path.dirname(__file__), "processes")


@override_settings(
    FLOW_PROCESSES_FINDERS=["resolwe.flow.finders.FileSystemProcessesFinder"]
)
@override_settings(FLOW_PROCESSES_DIRS=[PROCESSES_DIR])
@override_settings(FLOW_DESCRIPTORS_DIRS=[PROCESSES_DIR])
class ProcessRegisterTest(TestCase):
    def test_process_register_all(self):
        out, err = StringIO(), StringIO()
        call_command("register", stdout=out, stderr=err)
        self.assertIn("Inserted test-min", out.getvalue())
        self.assertIn(
            "Skip processor test-min: newer version installed", err.getvalue()
        )
        self.assertNotIn("setting is_active to True", out.getvalue())

        # Check that contributor gets all permissions.
        process = Process.objects.first()
        # TODO: this now returns EDIT which makes no sense but does not hurt either?
        self.assertEqual(len(process.get_permissions(self.admin)), 4)

        out, err = StringIO(), StringIO()
        call_command("register", stdout=out, stderr=err)
        self.assertIn("Skip processor test-min: same version installed", out.getvalue())
        self.assertIn(
            "Skip processor test-bloated: same version installed", out.getvalue()
        )
        self.assertIn(
            "Skip processor test-min: newer version installed", err.getvalue()
        )
        self.assertNotIn("setting is_active to True", out.getvalue())

        out, err = StringIO(), StringIO()
        call_command("register", force=True, stdout=out, stderr=err)
        self.assertIn("Updated test-min", out.getvalue())
        self.assertIn("Updated test-bloated", out.getvalue())
        self.assertIn(
            "Skip processor test-min: newer version installed", err.getvalue()
        )
        self.assertNotIn("setting is_active to True", out.getvalue())
        # The all_objects is important, since window function (used by the default
        # manager) is not compatible with the update statement.
        Process.all_objects.filter(slug="test-min").update(is_active=False)
        call_command("register", stdout=out, stderr=err)
        self.assertIn("Processor test-min: setting is_active to True", out.getvalue())

    def test_validation_of_defaults(self):
        out, err = StringIO(), StringIO()

        with self.settings(
            FLOW_PROCESSES_DIRS=[os.path.join(PROCESSES_DIR, "wrong_defaults")]
        ):
            call_command("register", stdout=out, stderr=err)

        self.assertIn("VALIDATION ERROR: Test Process Wrong Type", err.getvalue())
        self.assertIn("VALIDATION ERROR: Test Process Out of Range", err.getvalue())

    def test_inherit_perms(self):
        out, err = StringIO(), StringIO()

        with self.settings(
            FLOW_PROCESSES_DIRS=[os.path.join(PROCESSES_DIR, "first_version")]
        ):
            call_command("register", stdout=out, stderr=err)

        process = Process.objects.latest()
        process.set_permission(Permission.VIEW, self.user)

        out, err = StringIO(), StringIO()

        with self.settings(
            FLOW_PROCESSES_DIRS=[os.path.join(PROCESSES_DIR, "second_version")]
        ):
            call_command("register", stdout=out, stderr=err)

        process = Process.objects.latest()
        self.assertTrue(self.user.has_perm(Permission.VIEW, process))

    def test_retire(self):
        # No process should be in data base initially
        self.assertEqual(Process.objects.count(), 0)

        out, err = StringIO(), StringIO()
        # Register process test-proc, version 1.0.0 (first version of the process)
        # There is also version 2.0.0 of the same process
        with self.settings(
            FLOW_PROCESSES_DIRS=[os.path.join(PROCESSES_DIR, "first_version")]
        ):
            call_command("register", stdout=out, stderr=err)

        self.assertEqual(Process.objects.get(slug="test-proc").version, "1.0.0")

        call_command("register", stdout=out, stderr=err)

        # Check that all registered processes are active
        initial_processes = Process.objects.count()
        active_processes = Process.objects.filter(is_active=True).count()
        self.assertGreater(initial_processes, 0)
        self.assertEqual(initial_processes, active_processes)

        # Create data with one of the processes
        process = Process.objects.filter(slug="test-min").latest()
        Data.objects.create(
            name="Test min", process=process, contributor=self.contributor
        )

        # Nothing changes if register is called again in subfolder without --retire
        with self.settings(
            FLOW_PROCESSES_DIRS=[os.path.join(PROCESSES_DIR, "second_version")]
        ):
            call_command("register", stdout=out, stderr=err)

        self.assertEqual(initial_processes, Process.objects.count())
        self.assertEqual(
            initial_processes, Process.objects.filter(is_active=True).count()
        )

        # Check that retired processes without data are removed and
        # that retired processes with data are deactivated
        with self.settings(
            FLOW_PROCESSES_DIRS=[
                os.path.join(PROCESSES_DIR, "first_version"),
                os.path.join(PROCESSES_DIR, "second_version"),
            ]
        ):
            call_command("register", "--retire", stdout=out, stderr=err)

        # One process had data and is kept but inactive
        # Of the two process of the same slug only the latest remains
        self.assertEqual(Process.objects.count(), 2)
        # The process that is left in the schema path remains active and is
        # of the latest version
        self.assertEqual(Process.objects.filter(is_active=True).count(), 1)
        self.assertEqual(Process.objects.filter(is_active=True)[0].slug, "test-proc")
        self.assertEqual(Process.objects.filter(is_active=True)[0].version, "2.0.0")
        # The process with associated data is inactive
        self.assertEqual(Process.objects.filter(is_active=False).count(), 1)


@override_settings(
    FLOW_PROCESSES_FINDERS=["resolwe.flow.finders.FileSystemProcessesFinder"]
)
@override_settings(FLOW_PROCESSES_DIRS=[PROCESSES_DIR])
@override_settings(FLOW_DESCRIPTORS_DIRS=[PROCESSES_DIR])
class ProcessRegisterTestNoAdmin(DjangoTestCase):
    def test_process_register_no_admin(self):
        err = StringIO()
        self.assertRaises(SystemExit, call_command, "register", stderr=err)
        self.assertEqual("Admin does not exist: create a superuser\n", err.getvalue())


class ListDockerImagesTest(ProcessTestCase):
    def setUp(self):
        super().setUp()

        # Make sure the test processes are in the database
        self._register_schemas(processes_paths=[PROCESSES_DIR])

    def test_basic_list(self):
        # List Docker images and check if there's at least one
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", stdout=out, stderr=err)
        self.assertNotEqual("", out.getvalue())
        self.assertEqual("", err.getvalue())

    def test_basic_list_yaml(self):
        # List Docker images in YAML format and see if the output is valid YAML
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", format="yaml", stdout=out, stderr=err)
        self.assertNotEqual("", out.getvalue())
        self.assertEqual("", err.getvalue())
        imgs = yaml.safe_load(out.getvalue())
        self.assertTrue(isinstance(imgs, list))
        self.assertTrue(len(imgs) != 0)
        self.assertTrue(isinstance(imgs[0], dict))

    def test_invalid_format(self):
        # An unsupported format option should return an error
        self.assertRaises(
            CommandError, call_command, "list_docker_images", format="invalid"
        )

    def test_list_has_test_image(self):
        # The returned list must contain the image 'resolwe/test:base'
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", stdout=out, stderr=err)
        self.assertEqual("", err.getvalue())
        self.assertTrue("resolwe/test:base" in out.getvalue())

    def test_list_has_test_image_yaml(self):
        # The returned list must contain the image 'resolwe/test:base',
        # in the YAML output as well
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", format="yaml", stdout=out, stderr=err)
        self.assertEqual("", err.getvalue())
        imgs = yaml.safe_load(out.getvalue())
        self.assertTrue(isinstance(imgs, list))
        self.assertTrue(len(imgs) != 0)
        self.assertTrue(isinstance(imgs[0], dict))
        self.assertTrue(dict(name="resolwe/test", tag="base") in imgs)

    def test_list_is_sorted(self):
        # The returned list must be sorted alphabetically
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", stdout=out, stderr=err)
        self.assertEqual("", err.getvalue())
        self.assertNotEqual("", out.getvalue())
        lines = out.getvalue().split("\n")[:-1]
        self.assertEqual(lines, sorted(lines))

    def test_list_versions(self):
        # The returned list must contain only the Docker image from
        # the latest version of a given process
        out, err = StringIO(), StringIO()
        call_command("list_docker_images", stdout=out, stderr=err)
        self.assertEqual("", err.getvalue())
        self.assertIn("resolwe/test:versioning-2", out.getvalue())
        self.assertNotIn("resolwe/test:versioning-1", out.getvalue())



"""Python process runner utility."""

import inspect
import logging
import os
import sys
import tempfile
import types
from base64 import b64decode
from typing import Dict, Type

from communicator import communicator

# Id of the Data object we are processing.
DATA_ID = int(os.getenv("DATA_ID", "-1"))
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    assert communicator is not None

    # Get the python runtime.
    python_runtime = communicator.get_python_runtime()
    tmpdir = tempfile.TemporaryDirectory()
    python_runtime_filename = os.path.join(tmpdir.name, "python_runtime.zip")
    with open(python_runtime_filename, "wb") as python_runtime_handle:
        python_runtime_handle.write(b64decode(python_runtime))

    sys.path.insert(0, python_runtime_filename)
    # From this line on the Python process runtime is available.

    from resolwe.process.descriptor import ValidationError
    from resolwe.process.models import Data
    from resolwe.process.runtime import Process

    python_program = communicator.get_python_program()
    # Get the data object.
    data = Data(DATA_ID)

    module = types.ModuleType("python_program")
    exec(python_program, module.__dict__)

    # Mapping between the process slug and the class containing the process
    # definition.
    processes: Dict[str, Type[Process]] = {}
    for variable in dir(module):
        value = getattr(module, variable)
        if (
            value == Process
            or not inspect.isclass(value)
            or not issubclass(value, Process)
            or "_abstract" in value.__dict__
        ):
            continue
        processes[value._meta.metadata.slug] = value

    try:
        process = processes[data.process.slug]
    except KeyError:
        print("Found the following processes in module '{}':".format(processes))
        print("")
        for slug, process in processes.items():
            print("  {} ({})".format(slug, getattr(process, "name")))
        print("")

        print("ERROR: Unable to find process '{}'.".format(data.process.slug))
        sys.exit(1)

    # Start the process.
    try:
        process(data).start()
    except ValidationError as error:
        print("ERROR: Output field validation failed: {}".format(error.args[0]))
        sys.exit(1)
